-- Author:          Ernest Jamro
-- plik sluzy przede wszystkim do celow dydaktycznych
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset module (ROC)


entity OPB_MEM is
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	  C_REAL_DWIDTH : integer:= 8; -- real width of the data - the rst LSBs are always 0s
    C_SPEEDRATE : integer := 0 -- speed rate -- the lowest the value the better (start with 0), increase when the timing constrain condition is not satisfy
    );
  port (
    -- Global signals
    OPB_Clk : in std_logic;
    OPB_Rst : in std_logic; -- not used in the project
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);

    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic
    );
end entity OPB_MEM;


architecture opb_mem_arch of OPB_MEM is

function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
	for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "Address space is not a power of 2"
	severity failure;
	return -1;
end log2;

  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant awidth: integer:= log2(adr_space); -- memory address width
  constant ignore_adr: integer:= log2(c_real_dwidth/8); -- the LSBs of the address that are ignored
  
  component ROC -- reset on Configuration
  port (O : out std_logic);
  end component;

  component OPB_slave1
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	  C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
	   -- and the lower occupied area however the lower clock frequency is required. 
     -- Therefore to increase the clock frequency this parameter should be increased
	  -- user interface data and address width
    USE_SeqAddr : integer:= 1; -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
	  DWIDTH : integer:= 8; -- real width of the data - the rest LSBs are always 0s
	  AWIDTH : integer:= 8 -- user address width
    );
  port ( clk : in std_logic; -- clock signal
    arst: in std_logic; -- asynchronous reset (needed only for simulation)
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- user signals
	  sel: out std_logic; -- select - active when OPB_select and proper address
	  ack: in std_logic; -- acknowlenge - slave is ready for data transfer
	  rnw: out std_logic; -- read not write - same assert OPB_read not write
	  adr: out std_logic_vector(0 to awidth-1); -- address bus
	  drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	  dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	  be: out std_logic_vector(0 to dwidth/8 -1); -- byte enable 
	  seqAddr: out std_logic -- same as OPB_seqAddr
    );
  end component;

  component mem -- bram module included in the Virtex
    generic (AWidth: integer:= 9; -- Address Bus width
	  DWidth: integer:= 8); -- Data Bus Width
    port (clk, we: in std_logic;
	  adr: in std_logic_vector(awidth-1 downto 0);
	  din: in std_logic_vector(dwidth-1 downto 0);
	  dout: out std_logic_vector(dwidth-1 downto 0));
  end component; 

  signal sel, ack, rnw: std_logic; -- device is selected (OPB_select and device address)
  signal rd_ready: std_logic; -- device is ready for reading
  signal arst: std_logic; -- asynchronous reset generate just after FPGA configuration
  signal mem_we: std_logic_vector(0 to c_real_dwidth/8-1); -- memory write enable
  signal mem_dout: std_logic_vector(0 to c_real_dwidth-1); -- memory data out
  signal adr: std_logic_vector(0 to awidth-1);
  signal dwr, drd: std_logic_vector(0 to c_real_dwidth-1);
  signal be: std_logic_vector(0 to c_real_dwidth/8 -1);
 begin
  slave: OPB_slave1
  generic map( C_OPB_AWIDTH=> C_OPB_AWIDTH, C_OPB_DWIDTH=> C_OPB_DWIDTH,
    C_BASEADDR=> C_BASEADDR, C_HIGHADDR=> C_HIGHADDR, C_SPEEDRATE=> C_SPEEDRATE,
	  -- user interface data and address width
	  DWIDTH=> C_REAL_DWIDTH, AWIDTH=> awidth)
  port map(
    clk=> OPB_clk, arst=> arst,
    -- OPB signals
    OPB_ABus=> OPB_ABus, OPB_BE=> OPB_BE, OPB_RNW=> OPB_RNW,
    OPB_select=> OPB_select, OPB_seqAddr=> OPB_seqAddr,
    OPB_DBus=> OPB_DBus, Sl_DBus=> Sl_DBus,
    Sl_errAck=> Sl_errAck, Sl_retry=> Sl_retry,
    Sl_toutSup=> Sl_toutSup, Sl_xferAck=> Sl_xferAck,
	-- user signals
	sel=> sel, ack=> ack, rnw=> rnw, adr=> adr, 
	drd=> drd, dwr=> dwr, be=> be, seqAddr=> open);
	 
  -- rd_ready Flip-flop	(signal needed for only reading and indicates that the device presents proper data on Sl_DBus)
  rd_rdy: process(OPB_clk, OPB_rst) begin
	if opb_rst='1' then
	  rd_ready<= '0';
	elsif opb_clk='1' and opb_clk'event then
	  rd_ready<= sel and rnw and not rd_ready; -- rd_ready can be active only for one clock cycle
		   -- this is in the case when two reads are done in two consecutive clocks
	end if;
  end process;
  
  -- generate seperate BRAM module for each Byte (for byte enable logic)
  mem_be: for i in 0 to c_real_dwidth/8-1 generate
    mema: mem -- BRAM module (is build-in in FPGA)
      generic map (AWidth=> awidth-ignore_adr, DWidth=> 8)
      port map (clk=> opb_clk, we=> mem_we(i), 
	    adr=> adr(0 to awidth-ignore_adr-1),
	    din=> dwr(8*i to 8*i+7), dout=> drd(8*i to 8*i+7));

	mem_we(i)<= sel and not rnw and BE(i); -- seperate Byte Enable
  end generate;
  
  ack<= sel when RNW='0' -- no wait state for a write
	  else rd_ready; -- one wait state for a read

 assert c_real_dwidth<=c_opb_dwidth
   report "opb_bram error. c_real_dwidth must be greater than c_opb_dwidth"
   severity failure;

  res: ROC -- reset on Configuration
  port map (O=> arst);


end architecture opb_mem_arch;



